# Copyright (c) 2020 Stefan Neumann and Pauli Miettinen
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import numpy

class MisraGries:
	def __init__(self, maxNumCounters):
		self.maxNumCounters = maxNumCounters
		self.numVertices = 0
		self.counters = {}

	def addNumber(self, number):
		if number in self.counters:
			self.counters[number] += 1
		else:
			' need to add a new counter '
			numCounters = len(self.counters)
			if numCounters >= self.maxNumCounters:
				self.removeSmallestCounter()
			self.addCounterForNumber(number)

	def addNumbersFromBinaryVector(self, binaryVector):
		self.numVertices += 1
		for entry in binaryVector:
			self.addNumber(entry)

	def removeSmallestCounter(self):
		' find smallest counter '
		minCounter = float('inf')
		for key in self.counters:
			if self.counters[key] < minCounter:
				minCounter = self.counters[key]

		' decrease all counters by value of smallest counter '
		keysToRemove = []
		for key in self.counters:
			self.counters[key] -= minCounter
			if self.counters[key] == 0:
				keysToRemove.append(key)

		for key in keysToRemove:
			del self.counters[key]

	def addCounterForNumber(self, number):
		self.counters[number] = 1

	def addCounters(self, otherMG):
		self.numVertices += otherMG.numVertices

		for key in otherMG.counters:
			otherMGCount = otherMG.counters[key]

			if key in self.counters:
				self.counters[key] += otherMGCount
			else:
				self.counters[key] = otherMGCount

	def removeSuperfluousCounters(self):
		while len(self.counters) >= self.maxNumCounters:
			self.removeSmallestCounter()

